@@ -1,51 +1,110 @@
 from copy import deepcopy
+from collections import Counter
 import csv
 import datetime
 import decimal
 import io
 import json
-from typing import List, Optional
+from typing import List, Optional, Union
 
-from pydantic import validator, Extra
+from pydantic import validator, Extra, BaseModel
 from sqlalchemy import create_engine
 from sqlalchemy.sql import sqltypes
 
-from . import Database, DDL
+from . import Database, Columns, Hook
+
+from ..core.errors import Ok
 
 db_parameters = ["project", "credentials_path", "location", "dataset"]
 
 
-class BigqueryDDL(DDL):
-    partition: Optional[str]
-    cluster: Optional[List[str]]
+class Bigquery(Database):
+    class DDL(BaseModel):
+        class Properties(BaseModel):
+            partition: Optional[str]
+            cluster: Optional[List[str]]
 
-    class Config:
-        extra = Extra.forbid
+            class Config:
+                extra = Extra.forbid
 
-    @validator("cluster")
-    def validate_cluster(cls, v, values):
-        if len(values.get("columns")) > 0:
-            missing_columns = set(v) - set([c.name for c in values.get("columns")])
-            if len(missing_columns) > 0:
-                raise ValueError(
-                    f'Cluster contains columns not specified in the ddl: "{missing_columns}"'
+        columns: Optional[List[Columns]] = list()
+        properties: Optional[List[Properties]] = list()
+        post_hook: Optional[List[Hook]] = list()
+
+        class Config:
+            extra = Extra.forbid
+
+        @validator("columns", pre=True)
+        def transform_str_cols(cls, v, values):
+            if v is not None and isinstance(v, List):
+                return [{"name": c} if isinstance(c, str) else c for c in v]
+            else:
+                return v
+
+        @validator("columns")
+        def columns_unique(cls, v, values):
+            dupes = {k for k, v in Counter([e.name for e in v]).items() if v > 1}
+            if len(dupes) > 0:
+                raise ValueError(f"Duplicate columns: {','.join(dupes)}")
+            else:
+                return v
+
+        @validator("properties")
+        def validate_cluster(cls, v, values):
+            if len(values.get("columns")) > 0:
+                missing_columns = set(v) - set([c.name for c in values.get("columns")])
+                if len(missing_columns) > 0:
+                    raise ValueError(
+                        f'Cluster contains columns not specified in the ddl: "{missing_columns}"'
+                    )
+
+            return v
+
+        def get_ddl(self):
+            cols = self.columns
+            self.columns = []
+            for c in cols:
+                tests = []
+                for t in c.tests:
+                    if isinstance(t, str):
+                        tests.append({"type": t, "values": [], "execute": True})
+                    else:
+                        tests.append(
+                            {
+                                "type": t.name if t.name is not None else "values",
+                                "values": t.values if t.values is not None else [],
+                                "execute": t.execute,
+                            }
+                        )
+                self.columns.append(
+                    {
+                        "name": c.name,
+                        "description": c.description,
+                        "dst_name": c.dst_name,
+                        "type": c.type,
+                        "tests": tests,
+                    }
                 )
 
-        return v
+            props = self.properties
+            self.properties = []
+            for p in props:
+                self.properties.append(p.dict())
 
-    def get_ddl(self):
-        return {
-            "columns": [c.dict() for c in self.columns],
-            "indexes": {},
-            "permissions": self.permissions,
-            "partition": self.partition,
-            "cluster": self.cluster,
-            "primary_key": list(),
-        }
+            hook = self.post_hook
+            self.post_hook = []
+            for h in hook:
+                self.post_hook.append(h.dict())
 
+            res = {
+                "columns": self.columns,
+                "properties": self.properties,
+                "post_hook": self.post_hook,
+            }
 
-class Bigquery(Database):
-    ddl_validation_class = BigqueryDDL
+            return res
+
+    # ddl_validation_class = DDL
 
     project = None
     dataset = None
@@ -68,9 +127,90 @@ class Bigquery(Database):
 
         return create_engine(url, **settings)
 
-    def _introspect(self):
-        for schema in self._requested_objects.keys():
-            obj = [obj_name for obj_name in self._requested_objects[schema]]
+    def _construct_tests(self, columns, table, schema=None):
+        query = """
+                   SELECT val
+                        , col
+                        , cnt
+                        , type
+                     FROM (
+                """
+        template = self._jinja_test.get_template("standard_tests_bigquery.sql")
+        count_tests = 0
+        breakdown = []
+        for col in columns:
+            tests = col["tests"]
+            for t in tests:
+                breakdown.append(
+                    {
+                        "column": col["name"],
+                        "type": t["type"],
+                        "status": col[t["type"]],
+                        "execute": t["execute"],
+                    }
+                )
+                if col[t["type"]] is False and t["execute"]:
+                    count_tests += 1
+                    query += template.render(
+                        **{
+                            "table": table,
+                            "schema": schema,
+                            "name": col["name"]
+                            if not col["dst_name"]
+                            else col["dst_name"],
+                            "type": t["type"],
+                            "values": ", ".join(f"'{c}'" for c in t["values"]),
+                        },
+                    )
+            breakdown.append({"print": True})
+        breakdown = breakdown[:-1]
+        parts = query.splitlines()[:-2]
+        query = ""
+        for q in parts:
+            query += q.strip() + "\n"
+        query += ") AS t;"
+
+        if count_tests == 0:
+            return Ok(["", breakdown])
+
+        return Ok([query, breakdown])
+
+    def _format_properties(self, properties):
+
+        if properties["columns"]:
+            columns = []
+            for col in properties["columns"]:
+                entry = {
+                    "name": col["name"],
+                    "type": col["type"],
+                    "dst_name": col["dst_name"],
+                    "unique": False,
+                    "not_null": False,
+                    "values": False,
+                }
+                if "tests" in col:
+                    entry.update({"tests": col["tests"]})
+                    for t in col["tests"]:
+                        if (
+                            t["type"] != "values"
+                            and t["type"] != "unique"
+                            and col["type"]
+                        ):
+                            entry.update({t["type"]: True})
+                columns.append(entry)
+
+            properties["columns"] = columns
+        if properties["properties"]:
+            for pro in properties["properties"]:
+                for p in pro:
+                    if pro[p] is not None:
+                        properties[p] = pro[p]
+
+        return Ok(properties)
+
+    def _introspect(self, to_introspect):
+        for schema in to_introspect.keys():
+            obj = [obj_name for obj_name in to_introspect[schema]]
             if schema is None:
                 name = self.dataset
             else:
@@ -87,6 +227,9 @@ class Bigquery(Database):
                     """
             res = self.read_data(query)
 
+            if schema not in self._requested_objects:
+                self._requested_objects[schema] = dict()
+
             for obj in res:
                 obj_name = obj["table_name"]
                 if obj["table_type"] == "BASE TABLE":
@@ -174,6 +317,9 @@ class Bigquery(Database):
         job.result()
 
     def move_table(self, src_table, dst_table, src_schema=None, dst_schema=None, **ddl):
+
+        # ddl = self._format_properties(ddl).value
+
         full_src_table = (
             f"{src_schema + '.' if src_schema is not None else ''}{src_table}"
         )
@@ -192,6 +338,7 @@ class Bigquery(Database):
         replace=False,
         **ddl,
     ):
+
         full_name = fully_qualify(table, schema)
         if (
             schema in self._requested_objects
@@ -214,13 +361,19 @@ class Bigquery(Database):
             partition_column = None
             cluster_column = None
 
-        des_clustered = ddl["cluster"]
-        des_partitioned = ddl["partition"]
+        try:
+            des_clustered = ddl.get("cluster")
+        except:
+            des_clustered = ""
+        try:
+            des_partitioned = ddl.get("partition")
+        except:
+            des_partitioned = ""
 
         if des_clustered == cluster_column and des_partitioned == partition_column:
             drop = ""
         else:
-            drop = f"DROP TABLE IF EXISTS { table };"
+            drop = f"DROP TABLE IF EXISTS {schema}.{table};"
 
         template = self._jinja_env.get_template("create_table.sql")
         query = template.render(
