@@ -1,13 +1,16 @@
 from datetime import datetime
-from typing import Any, Dict, Optional
+from typing import Any, Dict, Optional, List
 
 from pydantic import BaseModel, Field, validator, Extra
 from sqlalchemy import or_, select, column
+from terminaltables import AsciiTable
 
 from ..core.errors import Err, Exc, Ok
 from ..database import Database
 from .sql import SqlTask
 
+# from .test import Columns
+
 
 class Source(BaseModel):
     supports_schemas: bool
@@ -64,12 +67,15 @@ class Destination(BaseModel):
 class Config(BaseModel):
     source: Source
     destination: Destination
-    ddl: Optional[Dict[str, Any]]
+    # ddl: Optional[Dict[str, Any]]
     delete_key: Optional[str]
     append: bool = False
     incremental_key: Optional[str]
     max_merge_rows: Optional[int]
     max_batch_rows: Optional[int]
+    columns: Optional[List[Dict[str, Any]]] = list()
+    table_properties: Optional[List[Dict[str, Any]]] = list()
+    post_hook: Optional[List[Dict[str, Any]]] = list()
 
     class Config:
         extra = Extra.forbid
@@ -102,7 +108,13 @@ class Config(BaseModel):
 
 
 class CopyTask(SqlTask):
-    def setup(self, **config):
+    def config(self, **config):  # noqa: C901
+        if "task_name" in self._config_input:
+            del self._config_input["task_name"]
+
+        if "columns" in config:
+            self._has_tests = True
+
         conn_names_list = [
             n for n, c in self.connections.items() if isinstance(c, Database)
         ]
@@ -154,42 +166,61 @@ class CopyTask(SqlTask):
             )
 
         try:
-            self.config = Config(**config)
+            self.task_config = Config(**config)
         except Exception as e:
             return Exc(e)
 
-        self.source_db = self.connections[self.config.source.db]
-        self.source_schema = self.config.source.db_schema
-        self.source_table = self.config.source.table
-        self.use_db_object(
-            self.source_table,
-            schema=self.source_schema,
-            db=self.source_db,
-            request_tmp=False,
+        # Setup sources
+        self.source_db = self.connections[self.task_config.source.db]
+        self.source_schema = self.task_config.source.db_schema
+        self.source_table = self.task_config.source.table
+
+        # Given copy's use case, we should take the table and schema as it comes
+        # TODO add some specification for using `src` so that it works for reverse ETL
+        # if self.source_schema is None:
+        #     self.src(self.source_table, connection=self.source_db)
+        # else:
+        #     self.src(
+        #         f"{self.source_schema}.{self.source_table}", connection=self.source_db
+        #     )
+
+        # Setup outputs
+        self.config_tmp_schema = (
+            self.task_config.destination.tmp_schema
+            or self.task_config.destination.db_schema
         )
+        self.config_schema = self.task_config.destination.db_schema
+        self.config_table = self.task_config.destination.table
+        self.config_tmp_table = f"sayn_tmp_{self.config_table}"
 
-        self.tmp_schema = (
-            self.config.destination.tmp_schema or self.config.destination.db_schema
-        )
-        self.schema = self.config.destination.db_schema
-        self.table = self.config.destination.table
-        self.tmp_table = f"sayn_tmp_{self.table}"
+        if self.config_schema is None:
+            self.schema = None
+            self.table = self.out(self.config_table, connection=self.target_db)
+        else:
+            obj = self.out(
+                f"{self.config_schema}.{self.config_table}", connection=self.target_db
+            )
+            self.schema = obj.split(".")[0]
+            self.table = obj.split(".")[1]
 
-        self.use_db_object(
-            self.table,
-            schema=self.schema,
-            tmp_schema=self.tmp_schema,
-            db=self.target_db,
-            request_tmp=True,
-        )
+        if self.config_tmp_schema is None:
+            self.tmp_schema = None
+            self.tmp_table = self.out(self.config_tmp_table, connection=self.target_db)
+        else:
+            obj = self.out(
+                f"{self.config_schema}.{self.config_tmp_table}",
+                connection=self.target_db,
+            )
+            self.tmp_schema = obj.split(".")[0]
+            self.tmp_table = obj.split(".")[1]
 
-        self.delete_key = self.config.delete_key
-        self.src_incremental_key = self.config.incremental_key
-        self.dst_incremental_key = self.config.incremental_key
-        self.max_merge_rows = self.config.max_merge_rows
-        self.max_batch_rows = self.config.max_batch_rows
+        self.delete_key = self.task_config.delete_key
+        self.src_incremental_key = self.task_config.incremental_key
+        self.dst_incremental_key = self.task_config.incremental_key
+        self.max_merge_rows = self.task_config.max_merge_rows
+        self.max_batch_rows = self.task_config.max_batch_rows
 
-        if self.config.append:
+        if self.task_config.append:
             self.mode = "append"
         elif self.dst_incremental_key is None:
             self.mode = "full"
@@ -198,19 +229,63 @@ class CopyTask(SqlTask):
 
         self.is_full_load = self.run_arguments["full_load"] or self.mode == "full"
 
-        result = self.target_db._validate_ddl(self.config.ddl)
+        result = self.target_db._validate_ddl(
+            self.task_config.columns,
+            self.task_config.table_properties,
+            self.task_config.post_hook,
+        )
         if result.is_ok:
-            self.ddl = result.value
+            self.columns = result.value
 
             # Check if the incremental_key in the destination needs renaming
-            if self.dst_incremental_key is not None and len(self.ddl["columns"]) > 0:
+            if (
+                self.dst_incremental_key is not None
+                and len(self.columns["columns"]) > 0
+            ):
                 columns_dict = {
-                    c["name"]: c["dst_name"] or c["name"] for c in self.ddl["columns"]
+                    c["name"]: c["dst_name"] or c["name"]
+                    for c in self.columns["columns"]
                 }
                 self.dst_incremental_key = columns_dict[self.src_incremental_key]
         else:
             return result
 
+        if self.run_arguments["command"] == "test":
+            result = self.target_db._construct_tests(
+                self.columns["columns"], self.table, self.schema
+            )
+            if result.is_err:
+                return result
+            else:
+                self.test_query = result.value[0]
+                self.test_breakdown = result.value[1]
+
+        return Ok()
+
+    def setup(self, needs_recompile):
+        if needs_recompile:
+            if self.config_schema is None:
+                self.table = self.out(self.config_table, connection=self.target_db)
+            else:
+                obj = self.out(
+                    f"{self.config_schema}.{self.config_table}",
+                    connection=self.target_db,
+                )
+                self.schema = obj.split(".")[0]
+                self.table = obj.split(".")[1]
+
+            if self.config_tmp_schema is None:
+                self.tmp_table = self.out(
+                    self.config_tmp_table, connection=self.target_db
+                )
+            else:
+                obj = self.out(
+                    f"{self.config_schema}.{self.config_tmp_table}",
+                    connection=self.target_db,
+                )
+                self.tmp_schema = obj.split(".")[0]
+                self.tmp_table = obj.split(".")[1]
+
         return Ok()
 
     def compile(self):
@@ -243,6 +318,37 @@ class CopyTask(SqlTask):
 
         return result
 
+    def test(self):
+        if self.test_query == "":
+            self.info(self.get_test_breakdown(self.test_breakdown))
+            return self.success()
+        else:
+            with self.step("Write Test Query"):
+                self.write_compilation_output(self.test_query, "test")
+
+            with self.step("Execute Test Query"):
+                result = self.default_db.read_data(self.test_query)
+
+                self.info(self.get_test_breakdown(self.test_breakdown))
+
+                if len(result) == 0:
+                    return self.success()
+                else:
+                    errout = "Test failed, summary:\n\n"
+                    errout += f"Total number of offending records: {sum([item['cnt'] for item in result])} \n"
+                    data = []
+                    data.append(
+                        ["Breach Count", "Prob. Value", "Test Type", "Failed Fields"]
+                    )
+
+                    for res in result:
+                        data.append([res["cnt"], res["val"], res["type"], res["col"]])
+                    table = AsciiTable(data)
+
+                    errinfo = f"You can find the compiled test query at compile/{self.group}/{self.name}_test.sql"
+
+                    return self.fail(errout + table.table + "\n\n" + errinfo)
+
     def execute(self, execute, debug, is_full_load, limit=None):
         # Introspect target
         self.target_table_exists = self.target_db._table_exists(self.table, self.schema)
@@ -268,13 +374,14 @@ class CopyTask(SqlTask):
             else:
                 get_data_query = result.value
 
-            create_ddl = {k: v for k, v in self.ddl.items() if k != "columns"}
+            create_ddl = {k: v for k, v in self.columns.items() if k != "columns"}
+
             if self.mode == "append":
-                create_ddl["columns"] = [c for c in self.ddl["columns"]] + [
+                create_ddl["columns"] = [c for c in self.columns["columns"]] + [
                     {"name": "_sayn_load_ts", "type": "TIMESTAMP"}
                 ]
             else:
-                create_ddl["columns"] = [c for c in self.ddl["columns"]]
+                create_ddl["columns"] = [c for c in self.columns["columns"]]
 
             query = self.target_db.create_table(
                 load_table, schema=load_schema, replace=True, **create_ddl
@@ -318,7 +425,7 @@ class CopyTask(SqlTask):
                 self.table,
                 src_schema=load_schema,
                 dst_schema=self.schema,
-                **self.ddl,
+                **self.columns,
             )
         elif final_step == "Merge Tables":
             query = self.target_db.merge_tables(
@@ -327,7 +434,7 @@ class CopyTask(SqlTask):
                 self.delete_key,
                 src_schema=load_schema,
                 dst_schema=self.schema,
-                **self.ddl,
+                **self.columns,
             )
         else:
             query = None
@@ -362,14 +469,14 @@ class CopyTask(SqlTask):
             )
         self.source_table_def = source_table_def
 
-        if len(self.ddl["columns"]) == 0:
+        if len(self.columns["columns"]) == 0:
             dst_table_def = None
             if not self.is_full_load:
                 dst_table_def = self.target_db._get_table(self.table, self.schema)
 
             if dst_table_def is not None:
                 # In incremental loads we use the destination table to determine the columns
-                self.ddl["columns"] = [
+                self.columns["columns"] = [
                     {
                         "name": c.name,
                         "type": c.type.compile(dialect=self.target_db.engine.dialect),
@@ -404,10 +511,10 @@ class CopyTask(SqlTask):
                         col_type = self.target_db._py2sqa(c.type.python_type)
                     except:
                         col_type = c.type.compile(self.target_db.engine.dialect)
-                    self.ddl["columns"].append({"name": c.name, "type": col_type})
+                    self.columns["columns"].append({"name": c.name, "type": col_type})
         else:
             # Fill up column types from the source table
-            for col in self.ddl["columns"]:
+            for col in self.columns["columns"]:
                 if col.get("name") not in self.source_table_def.columns:
                     return Err(
                         "database_error",
@@ -428,7 +535,7 @@ class CopyTask(SqlTask):
                             self.source_table_def.columns[col["name"]].type.python_type
                         )
 
-        for col in self.ddl["columns"]:
+        for col in self.columns["columns"]:
             col["src_name"] = col["name"]
             if col.get("dst_name") is not None:
                 col["name"] = col["dst_name"]
@@ -452,7 +559,7 @@ class CopyTask(SqlTask):
                 column(c["src_name"]).label(c["name"])
                 if c["src_name"] != c["name"]
                 else column(c["src_name"])
-                for c in self.ddl["columns"]
+                for c in self.columns["columns"]
             ],
             from_obj=self.source_table_def,
         )
